use anyhow::{anyhow, Result};

use std::{io, string::FromUtf8Error};

pub use serialport::{DataBits, FlowControl, Parity, SerialPort, StopBits};
pub use std::time::Duration;

pub enum SerialReadResult {
    Message(String),
    UtfConversion(FromUtf8Error),
    Timeout,
}

pub struct SerialPortSettings {
    pub baud_rate: u32,
    pub data_bits: DataBits,
    pub parity: Parity,
    pub stop_bits: StopBits,
    pub flow_control: FlowControl,
    pub timeout: Duration,
}

pub struct UsbId {
    pub vendor_id: u16,
    pub product_id: u16,
}

pub struct Port {
    serial_port: Box<dyn SerialPort>,
}

impl Port {
    pub fn open(usb_device: UsbId, settings: SerialPortSettings) -> Result<Self> {
        let port_path = Self::loop_usb_devices(usb_device)?;

        println!("found device at path {}", port_path);

        let port = serialport::new(port_path, settings.baud_rate)
            .data_bits(settings.data_bits)
            .parity(settings.parity)
            .stop_bits(settings.stop_bits)
            .flow_control(settings.flow_control)
            .timeout(settings.timeout)
            .open()?;

        Ok(Port { serial_port: port })
    }

    #[allow(unused)]
    pub fn read(&mut self) -> Result<SerialReadResult> {
        let mut buf: Vec<u8> = vec![0; 256];

        match self.serial_port.read(&mut buf[..]) {
            Ok(t) => match String::from_utf8(buf[0..t].to_vec()) {
                Ok(s) => Ok(SerialReadResult::Message(s)),
                Err(err) => Ok(SerialReadResult::UtfConversion(err)),
            },
            Err(ref e) if e.kind() == io::ErrorKind::TimedOut => Ok(SerialReadResult::Timeout),
            Err(err) => Err(anyhow!("failed reading serial port ({})", err)),
        }
    }

    pub fn write(&mut self, msg: &str) -> Result<()> {
        self.serial_port
            .write(msg.as_bytes())
            .map_err(|err| anyhow!("failed writing to serial port ({})", err))?;

        Ok(())
    }

    fn loop_usb_devices(usb_device: UsbId) -> Result<String> {
        for _ in 0..3 {
            if let Some(device) = Self::find_macroboard(&usb_device)? {
                return Ok(device);
            }

            println!(
                "no device found with {}:{}",
                usb_device.vendor_id, usb_device.product_id
            );

            std::thread::sleep(Duration::from_secs(2));
        }

        Err(anyhow!("no device found"))
    }

    fn find_macroboard(usb_device: &UsbId) -> Result<Option<String>> {
        let available_ports = serialport::available_ports()
            .map_err(|err| anyhow!("error querying serial ports ( {})", err))?;

        for available_port in available_ports.iter() {
            if let serialport::SerialPortType::UsbPort(usb_info) = &available_port.port_type {
                // check for the correct device
                if usb_info.vid == usb_device.vendor_id && usb_info.pid == usb_device.product_id {
                    return Ok(Some(available_port.port_name.clone()));
                }
            }
        }

        return Ok(None);
    }
}
